# SPDX-FileCopyrightText: 2023-present Ashok Cutkosky <ashok@cutkosky.com>
#
# SPDX-License-Identifier: Apache-2.0

from ._dynamic_scalar import (
    DynamicScalerState,
    dynamic_scale_tx,
    dynamic_scale_value_and_grad,
    dynamic_scale_grad,
)
from ._amp import (
    amp,
    amp_stop,
    default_amp_policy,
    use_original_precision,
    use_compute_precision,
)
